#include <msp430.h> 
#include <msp430g2553.h>
#include <stdbool.h>
#include <bq769x0_I2C_G2553.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
/*
 * main.c
 */

//RegisterGroup Registers;

RegisterGroup Upper_Registers, Bottom_Registers, Registers;

const unsigned int OVPThreshold = 3700;
const unsigned int UVPThreshold = 2500;
const unsigned char SCDDelay = SCD_DELAY_400us;
const unsigned char SCDThresh = SCD_THRESH_44mV_22mV;
const unsigned char OCDDelay = OCD_DELAY_320ms;
const unsigned char OCDThresh = OCD_THRESH_22mV_11mV;
const unsigned char OVDelay = OV_DELAY_1s;
const unsigned char UVDelay = UV_DELAY_1s;
const unsigned char UARTTimeOut = 2;
const int SleepDelay = 20;
const unsigned char ShutdownCommand[3] = {0, 0x10, 0};
const unsigned char SleepCommand[3] = {0, 0x11, 0};

unsigned int CellVoltage[20];
float Gain = 0;
int iGain = 0;

unsigned char Semaphore = 0;

//Variables for UART communication
unsigned char CommandStatus = COMMAND_IDLE;
CommandFrame Command;
unsigned char ReceiveBuffer[33];
unsigned char ReceiveBufferIndex = 0;
unsigned char *SendBuffer = NULL;
unsigned char SendBufferIndex = 0;
unsigned char SendByteNumber = 0;
unsigned char ResponseCode = 0;
unsigned char ParseBufferIndex = 0;
unsigned char BufferFull = 0;
unsigned char DataCounterInFrame = 0;
unsigned char FrameLength = 0;
unsigned char TickCounter = 0;
unsigned char UARTCounter = 0;

//Variables for Timer0
int SleepTimer = 0;
unsigned char PowerMode = POWER_MODE_ACTIVE;

void Delay10mS()
{
	unsigned int i;

	for (i = 0; i < 14708; i++)
	{}

	WDT_CLEAR;
	//P1OUT ^= BIT3;

}

void ClockInitialise()
{
	DCOCTL = 0;		//set the lowest DCOx and MODx settings

	DCOCTL = CALDCO_16MHZ;
	BCSCTL1 |= CALBC1_16MHZ;	//select 16MHz DCO
	BCSCTL1 |= DIVA_3;			//select divider for ACLK, divide by 8 for ACLK;
	BCSCTL1 &= ~XTS;			//To use VLO, low frequency mode

	BCSCTL2 = 0x0;				//select DCOCLK for MCLK; MCLK = DCOCLK; SMCLK = DCOCLK; divider for SMCLK = 1; use internal resistor;

	BCSCTL3 |= LFXT1S_2;		//To use VLO, 12kHz?
}

void TimerAInitialise()
{
	TACTL = TASSEL0 + ID1 + TACLR;		//diviver = 4, clock source ACLK, VLOCLK, 12kHz
	TACCR0 = 375;						//divide 12000Hz to 1Hz by 12000/4/1/8 * 375, divider 1 comes from the ratio from VLOCLK to ACLK
}

void TimerAStart()
{
	TACTL |= TACLR;						//clear the TAR

	while(TACTL & TACLR)
	{}

	TACTL |= MC0;						//the timer is halted

	TACCTL0 |= CCIE;					//Interrupt enable
}

void Port1Init()
{
	//set P2.3 to shutdown control function
	P1OUT |= BIT3;			//set resistor to pull up
	P1REN |= BIT3;			//enable pull up/down resistor
	P1SEL &= ~BIT3;
	P1SEL2 &= ~BIT3;		//select the basic IO function
	P1DIR &= ~BIT3;			//select the input function
	//P1IES |= BIT3;			//set the IFG on high to low transition, this is for UART_CAN transaction detection
	//P1IE |= BIT3 + BIT4 + BIT5;		//BIT4 and BIT5 triggered on low to high transition, Bit3 triggered on High to Low transition
	P1IE |= BIT4 + BIT5;				//temperorily disable P1.3 interrupt
}

void Port3Init()
{
	P3SEL &= ~(BIT6 + BIT7);
	P3SEL2 &= ~(BIT6 + BIT7);
	P3DIR |= BIT6 + BIT7;						//for debug purpose only
	ADC10AE0 &= ~(BIT6 + BIT7);

}

void Port4Init()
{
	P4SEL = 0;
	P4SEL2 = 0;
	P4REN &= BIT2 + BIT6 + BIT7;
	P4DIR |= 0xff;						//P4 all set to GPIO output
	CAPD &= ~(BIT2 + BIT6 + BIT7);
	P4OUT = BIT2;

}

void ADC10Init()
{
	ADC10AE0 = BIT0;
	ADC10AE1 = 0;						//Only set A0 as Input Channel

	ADC10CTL0 = SREF0 + ADC10SHT1 + ADC10SHT0 + ADC10SR + REFBURST + REF2_5V; //use internal reference, 64 ADCCLOCKs for sample and hold time,
	//50ksps sample rate, Reference Output OFF, REFOUT = 0, referenc on only upon sample and conversions, MSC = 0, requires a rising edge on SHI to triger sample and conversion
	//Select 2.5v as reference votage;

	ADC10CTL1 &= ADC10BUSY;				//INCH = 0, select A0 for input Channel, SHSx = 0, use ADC10SC bit as sample and conversion control bit, ADC10DF = 0, use straight binary format
	//ISSH = 0, Sample input signal is not inverted, ADC10DIVx = 0, ADC Clock Divider = 1, ADC10SSELx = 0, select ADC10OSC(5MHz), CONSEQx = 0, use single channel singel conversion

	ADC10DTC1 = 0;						//disable DTC
}

void PullDownSMBus()
{
	P1SEL &= ~(BIT6 + BIT7);
	P1SEL2 &= ~(BIT6 + BIT7);

	P1DIR |= (BIT6 + BIT7);	//set to output
	P1OUT &= ~(BIT6 + BIT7);	//pull down
}
/*
void ReviveSMBus()
{
	P1SEL |= BIT6 + BIT7;                     // Assign I2C pins to USCI_B0, P1.6 for SCL and P1.7 for SDA
	P1SEL2|= BIT6 + BIT7;                     // Assign I2C pins to USCI_B0, P1.6 for SCL and P1.7 for SDA
}
*/
/*
void SetI2CClockToACLK()
{
	UCB0BR0 = 1;
	UCB0BR1 = 0;
	UCB0CTL1 = UCSSEL_1;		//Select ACLK
}

void SetI2CClockToSMCLK()
{
	UCB0BR0 = 200;
	UCB0BR1 = 0;
	UCB0CTL1 = UCSSEL_2;		//Select SMCLK
}
*/
void I2CInitialise_G2955()
{
	P3SEL |= BIT1 + BIT2;                     // Assign I2C pins to USCI_B0, P3.1 for SCL and P3.2 for SDA
	P3SEL2 &= ~(BIT1 + BIT2);                     // Assign I2C pins to USCI_B0, P3.1 for SCL and P3.2 for SDA

	UCB0CTL1 |= UCSWRST;                      // Enable SW reset, hold USCI logic in reset state
	UCB0CTL0 = UCMODE_3 + UCSYNC;			//set to I2C mode, sync=1
	UCB0BR0 = 200;							//80kHz clock rate
	UCB0BR1 = 0;

	UCB0I2CIE = 0;
    IE2 &= ~(UCB0TXIE + UCB0RXIE);	//disable interrupts

    UCB0CTL1 |= UCSSEL_2;
    UCB0CTL1 &= ~UCSWRST;
}

int I2CSendByte(unsigned char I2CSlaveAddress, unsigned char data)
{
	unsigned long int DelayCounter = 0;

	UCB0CTL0 |= UCMST;
	UCB0I2CSA = I2CSlaveAddress;

	 UCB0CTL1 |= UCTR; //data in transmit direction
	 UCB0CTL1 |= UCTXSTT; //Generate Start Condition
	 	 	 	 	 	  //Send Start Byte

	 while(!(IFG2 & UCB0TXIFG))  //if UCB0TXIFG != 0, wait here
	 {
		 DelayCounter ++;
		 if (DelayCounter >= DELAY_LIMIT)
			 break;
	 }

	 if (DelayCounter >= DELAY_LIMIT)
		 return -1;

	 UCB0TXBUF = data;				// send the data

	 DelayCounter = 0;

	 while(DelayCounter < DELAY_LIMIT && !(IFG2 & UCB0TXIFG))
	 {
		 DelayCounter++;
	 }

	 if (DelayCounter >= DELAY_LIMIT)
		 return -1;

	 UCB0CTL1 |= UCTXSTP;			//send stop bit

	 DelayCounter = 0;

	 while(DelayCounter < DELAY_LIMIT && (UCB0CTL1 & UCTXSTP))
	 {
		 DelayCounter++;
	 }

	 if (DelayCounter >= DELAY_LIMIT)	//check if NACK condition occurred
		 return -1;
	 else
		 return 0;

}

int I2CSendBytes(unsigned char I2CSlaveAddress, unsigned char *DataBuffer, unsigned int ByteCount, unsigned int *SentByte)
{
	unsigned long int DelayCounter = 0;
    unsigned int NumberOfBytesSent = 0;
    unsigned char *DataPointer;

	UCB0CTL0 |= UCMST;
	UCB0I2CSA = I2CSlaveAddress;

     DataPointer = DataBuffer;

	 UCB0CTL1 |= UCTR; //data in transmit direction
	 UCB0CTL1 |= UCTXSTT; //Generate Start Condition
	 	 	 	 	 	  //Send Start Byte

	 while(!(IFG2 & UCB0TXIFG))  //if UCTXSTT != 0, wait here
	 {
		 DelayCounter ++;
		 if (DelayCounter > DELAY_LIMIT)
			 break;
	 }

	 if (DelayCounter >= DELAY_LIMIT)   //check if NACK condition occurred
	 {
		 *SentByte = NumberOfBytesSent;
		 UCB0CTL1 |= UCTXSTP;
		 return -1;
	 }

	 for(NumberOfBytesSent = 0; NumberOfBytesSent < ByteCount; NumberOfBytesSent++)
	 {
		 UCB0TXBUF= *DataPointer;

		 DelayCounter = 0;

		 while(DelayCounter < DELAY_LIMIT && (!(IFG2 & UCB0TXIFG) || (UCB0CTL1 & UCTXSTT)))	//check if the byte has been sent
		 {
			 DelayCounter++;
		 }

		 if (DelayCounter >= DELAY_LIMIT)	//check if NACK condition occurred
		 {
			 *SentByte = NumberOfBytesSent;
			 UCB0CTL1 |= UCTXSTP;				//send stop condition
			 return -1;
		 }

		 DataPointer++;
	 }

	 IFG2 &= ~UCB0TXIFG;
	 UCB0CTL1 |= UCTXSTP;		//send stop bit

	 DelayCounter = 0;

	 while(DelayCounter < DELAY_LIMIT && ((UCB0CTL1 & UCTXSTP)))
	 {
		 DelayCounter++;
	 }

	 *SentByte =  NumberOfBytesSent;

	 if (DelayCounter >= DELAY_LIMIT)	//check if NACK condition occurred
	 {
		 UCB0CTL1 |= UCSWRST;
		 return -1;
	 }
	 else
		 return 0;
}

int I2CWriteRegisterByte(unsigned char I2CSlaveAddress, unsigned char Register, unsigned char Data)
{
	unsigned char DataBuffer[2];
	unsigned int SentByte = 0;


	DataBuffer[0] = Register;
	DataBuffer[1] = Data;

	return(I2CSendBytes(I2CSlaveAddress, DataBuffer, 2, &SentByte));
}

int I2CWriteRegisterByteWithCRC(unsigned char I2CSlaveAddress, unsigned char Register, unsigned char Data)
{
	unsigned char DataBuffer[4];
	unsigned int SentByte = 0;

    DataBuffer[0] = I2CSlaveAddress << 1;
	DataBuffer[1] = Register;
	DataBuffer[2] = Data;
	DataBuffer[3] = CRC8(DataBuffer, 3, CRC_KEY);

	return(I2CSendBytes(I2CSlaveAddress, DataBuffer + 1, 3, &SentByte));
}

int I2CWriteRegisterWordWithCRC(unsigned char I2CSlaveAddress, unsigned char Register, unsigned int Data)
{
	unsigned char DataBuffer[6];
	unsigned int SentByte = 0;

    DataBuffer[0] = I2CSlaveAddress << 1;
	DataBuffer[1] = Register;
	DataBuffer[2] = LOW_BYTE(Data);
	DataBuffer[3] = CRC8(DataBuffer, 3, CRC_KEY);
	DataBuffer[4] = HIGH_BYTE(Data);
	DataBuffer[5] = CRC8(DataBuffer + 4, 1, CRC_KEY);

	return(I2CSendBytes(I2CSlaveAddress, DataBuffer + 1, 5, &SentByte));
}

int I2CWriteBlockWithCRC(unsigned char I2CSlaveAddress, unsigned char StartAddress, unsigned char *Buffer, unsigned char Length)
{
	unsigned char *BufferCRC, *Pointer;
	int i;
	unsigned int SentByte = 0;
	int result;

	BufferCRC = (unsigned char*)malloc(2*Length + 2);
	if (NULL == BufferCRC)
		return -1;

	Pointer = BufferCRC;
	*Pointer = I2CSlaveAddress << 1;
	Pointer++;
	*Pointer = StartAddress;
	Pointer++;
	*Pointer = *Buffer;
	Pointer++;
	*Pointer = CRC8(BufferCRC, 3, CRC_KEY);

	for(i = 1; i < Length; i++)
	{
        Pointer++;
        Buffer++;
        *Pointer = *Buffer;
		*(Pointer + 1) = CRC8(Pointer, 1, CRC_KEY);
		Pointer++;
	}

	result = I2CSendBytes(I2CSlaveAddress, BufferCRC + 1, 2*Length + 1, &SentByte);

	free(BufferCRC);
	BufferCRC = NULL;

	return result;
}

int I2CWriteRegisterWord(unsigned char I2CSlaveAddress, unsigned char Register, unsigned int Data)
{
	unsigned char DataBuffer[3];
	unsigned int SentByte = 0;

	DataBuffer[0] = Register;
	DataBuffer[1] = LOWBYTE(Data);
	DataBuffer[2] = HIGHBYTE(Data);

	return(I2CSendBytes(I2CSlaveAddress, DataBuffer, 3, &SentByte));
}

int I2CReadBytes(unsigned char I2CSlaveAddress, unsigned char *DataBuffer, unsigned int ExpectedByteNumber, unsigned int *NumberOfReceivedBytes)
{
	unsigned long int DelayCounter = 0;
    unsigned char *DataPointer;
    unsigned int *NumberOfReceivedBytesPointer;

    NumberOfReceivedBytesPointer = NumberOfReceivedBytes;
    *NumberOfReceivedBytesPointer = 0;

    UCB0CTL0 |= UCMST;
    DataPointer = DataBuffer;
	UCB0I2CSA = I2CSlaveAddress;

	UCB0CTL1 &= ~(UCTR); //data in receive direction

	UCB0CTL1 |= UCTXSTT; //Generate Start Condition

	 while((UCB0CTL1 & UCTXSTT)
			 )  //if UCTXSTT != 0, wait here
	 {
		 DelayCounter ++;
		 if (DelayCounter >= DELAY_LIMIT)
			 break;
	 }

	 if (DelayCounter >= DELAY_LIMIT || UCB0STAT & UCNACKIFG)   //check if NACK condition occurred
		 return -1;

	 for(*NumberOfReceivedBytesPointer = 0; *NumberOfReceivedBytesPointer < ExpectedByteNumber; (*NumberOfReceivedBytesPointer)++)
	 {
		 if(*NumberOfReceivedBytesPointer + 1 == ExpectedByteNumber)
			 UCB0CTL1 |= UCTXSTP;

		 DelayCounter = 0;

		 while(DelayCounter < DELAY_LIMIT && !(IFG2 & UCB0RXIFG))
		 {
			 DelayCounter++;
		 }

		 if(DelayCounter == DELAY_LIMIT)
		 {
			 UCB0CTL1 |= UCSWRST;   //if I2C overtime condition occurred, reset I2C engine
			 return -1;
		 }

		 *DataPointer = UCB0RXBUF;

		 DataPointer++;
	 }

	 DelayCounter = 0;
	 while(DelayCounter < DELAY_LIMIT && (UCB0CTL1 & UCTXSTP))
	 {
		 DelayCounter++;
	 }

	 if(DelayCounter >= DELAY_LIMIT)
	 {
		 UCB0CTL1 |= UCSWRST;
		 return -1;
	 }

	 return 0;

}

int I2CReadRegisterByte(unsigned char I2CSlaveAddress, unsigned char Register, unsigned char *Data)
{
	unsigned char TargetRegister = Register;
	unsigned int SentByte = 0;
	unsigned int ReadDataCount = 0;
    int ReadStatus = 0;
    int WriteStatus = 0;

	WriteStatus = I2CSendBytes(I2CSlaveAddress, &TargetRegister, 1, &SentByte);

	ReadStatus = I2CReadBytes(I2CSlaveAddress, Data, 1, &ReadDataCount);

	if (ReadStatus != 0 || WriteStatus != 0)
	{
		return -1;
	}

	return 0;
}

int I2CReadBlock(unsigned char I2CSlaveAddress, unsigned char StartRegisterAddress, unsigned char *Buffer, unsigned int BlockSize, unsigned int *NumberOfBytes)
{
	unsigned char TargetRegister = StartRegisterAddress;
	unsigned int SentByte = 0;
	int ReadStatus = 0;
	int WriteStatus = 0;

	WriteStatus = I2CSendBytes(I2CSlaveAddress, &TargetRegister, 1, &SentByte);

	ReadStatus = I2CReadBytes(I2CSlaveAddress, Buffer, BlockSize, NumberOfBytes);

	if(ReadStatus != 0 || WriteStatus != 0)
	{
		return -1;
	}

	return 0;
}

unsigned char CRC8(unsigned char *ptr, unsigned char len,unsigned char key)
{
	unsigned char i;
	unsigned char crc=0;
	while(len--!=0)
	{
		for(i=0x80; i!=0; i/=2)
		{
			if((crc & 0x80) != 0)
			{
				crc *= 2;
				crc ^= key;
			}
			else
				crc *= 2;

			if((*ptr & i)!=0)
				crc ^= key;
		}
		ptr++;
	}
	return(crc);
}

int I2CReadRegisterByteWithCRC(unsigned char I2CSlaveAddress, unsigned char Register, unsigned char *Data)
{
	unsigned char TargetRegister = Register;
	unsigned int SentByte = 0;
	unsigned char ReadData[2];
	unsigned int ReadDataCount = 0;
	unsigned char CRCInput[2];
	unsigned char CRC = 0;
    int ReadStatus = 0;
    int WriteStatus = 0;

	WriteStatus = I2CSendBytes(I2CSlaveAddress, &TargetRegister, 1, &SentByte);

	ReadStatus = I2CReadBytes(I2CSlaveAddress, ReadData, 2, &ReadDataCount);

	if (ReadStatus != 0 || WriteStatus != 0)
	{
		return -1;
	}

	CRCInput[0] = (I2CSlaveAddress << 1) + 1;
	CRCInput[1] = ReadData[0];

	CRC = CRC8(CRCInput, 2, CRC_KEY);

	if (CRC != ReadData[1])
		return -1;

	*Data = ReadData[0];
	return 0;
}

int I2CReadRegisterWordWithCRC(unsigned char I2CSlaveAddress, unsigned char Register, unsigned int *Data)
{
	unsigned char TargetRegister = Register;
	unsigned int SentByte = 0;
	unsigned char ReadData[4];
	unsigned int ReadDataCount = 0;
	unsigned char CRCInput[2];
	unsigned char CRC = 0;
    int ReadStatus = 0;
    int WriteStatus = 0;

	WriteStatus = I2CSendBytes(I2CSlaveAddress, &TargetRegister, 1, &SentByte);

	ReadStatus = I2CReadBytes(I2CSlaveAddress, ReadData, 4, &ReadDataCount);

	if (ReadStatus != 0 || WriteStatus != 0)
	{
		return -1;
	}

	CRCInput[0] = (I2CSlaveAddress << 1) + 1;
	CRCInput[1] = ReadData[0];

	CRC = CRC8(CRCInput, 2, CRC_KEY);

	if (CRC != ReadData[1])
		return -1;

	CRC = CRC8(ReadData + 2, 1, CRC_KEY);

	if (CRC != ReadData[3])
		return -1;

	*Data = ReadData[0];

	*Data = (*Data << 8) + ReadData[2];

	return 0;
}

int I2CReadBlockWithCRC(unsigned char I2CSlaveAddress, unsigned char Register, unsigned char *Buffer, unsigned char Length)
{
	unsigned char TargetRegister = Register;
	unsigned int SentByte = 0;
	unsigned char *ReadData = NULL, *StartData = NULL;
	unsigned int ReadDataCount = 0;
	unsigned char CRCInput[2];
	unsigned char CRC = 0;
    int ReadStatus = 0;
    int WriteStatus = 0;
    int i;

    StartData = (unsigned char *)malloc(2 * Length);

    if (NULL == StartData)
    	return -1;

    ReadData = StartData;

	WriteStatus = I2CSendBytes(I2CSlaveAddress, &TargetRegister, 1, &SentByte);

	ReadStatus = I2CReadBytes(I2CSlaveAddress, ReadData, 2 * Length, &ReadDataCount);

	if (ReadStatus != 0 || WriteStatus != 0)
	{
		free(StartData);
		StartData = NULL;

		return -1;
	}

	CRCInput[0] = (I2CSlaveAddress << 1) + 1;
	CRCInput[1] = *ReadData;

	CRC = CRC8(CRCInput, 2, CRC_KEY);

	ReadData++;
	if (CRC != *ReadData)
	{
		free(StartData);
		StartData = NULL;
		return -1;
	}
	else
		*Buffer = *(ReadData - 1);

	for(i = 1; i < Length; i++)
	{
		ReadData++;
		CRC = CRC8(ReadData, 1, CRC_KEY);
		ReadData++;
		Buffer++;

		if (CRC != *ReadData)
		{
			free(StartData);
			StartData = NULL;

			return -1;
		}
		else
			*Buffer = *(ReadData - 1);
	}

	free(StartData);
	StartData = NULL;

	return 0;
}

int GetADCGainOffset()
{
	int result;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, ADCGAIN1, &(Registers.ADCGain1.ADCGain1Byte));
	result = I2CReadRegisterByteWithCRC(BQMAXIMO, ADCGAIN2, &(Registers.ADCGain2.ADCGain2Byte));
	result = I2CReadRegisterByteWithCRC(BQMAXIMO, ADCOFFSET, &(Registers.ADCOffset));

	return result;
}

int ConfigureBqMaximo()
{
	int result = 0;
	unsigned char bqMaximoProtectionConfig[5];

	result = I2CWriteBlockWithCRC(BQMAXIMO, PROTECT1, &(Registers.Protect1.Protect1Byte), 5);

	result = I2CReadBlockWithCRC(BQMAXIMO, PROTECT1, bqMaximoProtectionConfig, 5);

	if(bqMaximoProtectionConfig[0] != Registers.Protect1.Protect1Byte
			|| bqMaximoProtectionConfig[1] != Registers.Protect2.Protect2Byte
			|| bqMaximoProtectionConfig[2] != Registers.Protect3.Protect3Byte
			|| bqMaximoProtectionConfig[3] != Registers.OVTrip
			|| bqMaximoProtectionConfig[4] != Registers.UVTrip)
	{
		result = -1;
	}

	return result;
}

int CheckBqMaximoConfiguration()
{
	int result = 0;
	unsigned char bqMaximoProtectionConfig[5];

	result = I2CReadBlockWithCRC(BQMAXIMO, PROTECT1, bqMaximoProtectionConfig, 5);

	if(bqMaximoProtectionConfig[0] != Registers.Protect1.Protect1Byte
			|| bqMaximoProtectionConfig[1] != Registers.Protect2.Protect2Byte
			|| bqMaximoProtectionConfig[2] != Registers.Protect3.Protect3Byte
			|| bqMaximoProtectionConfig[3] != Registers.OVTrip
			|| bqMaximoProtectionConfig[4] != Registers.UVTrip)
	{
		result = -1;
	}

	return result;
}

int CheckBqMaximoConfiguration_20S(unsigned char Device)
{
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	switch(Device)
	{
		case(BOTTOM_DEVICE):
				Registers = Bottom_Registers;
				break;
		case(UPPER_DEVICE):
				Registers = Upper_Registers;
				break;
		default:
			return GENERAL_CALL_FAIL;
	}

	result = CheckBqMaximoConfiguration();

	return result;
}

int InitialisebqMaximo()
{
	int result = 0;

	Registers.Protect1.Protect1Bit.SCD_DELAY = SCDDelay;
	Registers.Protect1.Protect1Bit.SCD_THRESH = SCDThresh;
	Registers.Protect2.Protect2Bit.OCD_DELAY = OCDDelay;
	Registers.Protect2.Protect2Bit.OCD_THRESH = OCDThresh;
	Registers.Protect3.Protect3Bit.OV_DELAY = OVDelay;
	Registers.Protect3.Protect3Bit.UV_DELAY = UVDelay;

	result = GetADCGainOffset();

	Gain = (365 + ((Registers.ADCGain1.ADCGain1Byte & 0x0C) << 1) + ((Registers.ADCGain2.ADCGain2Byte & 0xE0)>> 5)) / 1000.0;
	iGain = 365 + ((Registers.ADCGain1.ADCGain1Byte & 0x0C) << 1) + ((Registers.ADCGain2.ADCGain2Byte & 0xE0)>> 5);

    Registers.OVTrip = (unsigned char)((((unsigned short)((OVPThreshold - Registers.ADCOffset)/Gain + 0.5) - OV_THRESH_BASE) >> 4) & 0xFF);
    Registers.UVTrip = (unsigned char)((((unsigned short)((UVPThreshold - Registers.ADCOffset)/Gain + 0.5) - UV_THRESH_BASE) >> 4) & 0xFF);

    result = ConfigureBqMaximo();

    return result;
}

int InitialisebqMaximo_20S(unsigned char Device)
{
	int result = 0;

	if (Device != BOTTOM_DEVICE && Device != UPPER_DEVICE)
		return GENERAL_CALL_FAIL;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	result = InitialisebqMaximo();

	if (-1 == result)
		return result;

	if (BOTTOM_DEVICE == Device)
		Bottom_Registers = Registers;

	if (UPPER_DEVICE == Device)
		Upper_Registers = Registers;

	return 0;
}

int UpdateVoltageFromBqMaximo()
{
	int Result = 0, i = 0;
	unsigned char *pRawADCData = NULL;
	unsigned int iTemp = 0;
	unsigned long lTemp = 0;

	Result = I2CReadBlockWithCRC(BQMAXIMO, \
			VC1_HI_BYTE, \
			&(Registers.VCell1.VCell1Byte.VC1_HI), \
			30);

	pRawADCData = &Registers.VCell1.VCell1Byte.VC1_HI;
	for (i = 0; i < 15; i++)
	{
		iTemp = (unsigned int)(*pRawADCData << 8) + *(pRawADCData + 1);
		lTemp = ((unsigned long)iTemp * iGain)/1000;
		lTemp += Registers.ADCOffset;
		CellVoltage[i] = lTemp;
		pRawADCData += 2;
	}

	return Result;
}

int UpdateVoltageFromUpperBq769x0()
{
	int Result = 0, i = 0;
	unsigned char *pRawADCData = NULL;
	unsigned int iTemp = 0;
	unsigned long lTemp = 0;

	ENABLE_UPPER_BQ769x0;

	Result = I2CReadBlockWithCRC(BQMAXIMO, \
			VC1_HI_BYTE, \
			&(Registers.VCell1.VCell1Byte.VC1_HI), \
			30);

	pRawADCData = &Registers.VCell1.VCell1Byte.VC1_HI;
	for (i = 0; i < 10; i++)
	{
		iTemp = (unsigned int)(*pRawADCData << 8) + *(pRawADCData + 1);
		lTemp = ((unsigned long)iTemp * iGain)/1000;
		lTemp += Registers.ADCOffset;
		CellVoltage[i] = lTemp;
		pRawADCData += 2;
	}

	return Result;
}

int UpdateVoltageFromBottomBq769x0()
{
	int Result = 0, i = 0;
	unsigned char *pRawADCData = NULL;
	unsigned int iTemp = 0;
	unsigned long lTemp = 0;

	ENABLE_BOTTOM_BQ769x0;

	Result = I2CReadBlockWithCRC(BQMAXIMO, \
			VC1_HI_BYTE, \
			&(Registers.VCell1.VCell1Byte.VC1_HI), \
			30);

	pRawADCData = &Registers.VCell1.VCell1Byte.VC1_HI;
	for (i = 10; i < 20; i++)
	{
		iTemp = (unsigned int)(*pRawADCData << 8) + *(pRawADCData + 1);
		lTemp = ((unsigned long)iTemp * iGain)/1000;
		lTemp += Registers.ADCOffset;
		CellVoltage[i] = lTemp;
		pRawADCData += 2;
	}

	return Result;
}
void UARTInitialise_G2955()
{
	P3SEL |= BIT4 + BIT5;
	P3SEL2 &= ~(BIT4 + BIT5);

	UCA0CTL1 |= 1;			//Reset UART engine
	UCA0CTL0 |= UCMODE_1;    //Select idle-line mode, Parity disable, 8-bit data, one stop bit, asynchronous mode
	UCA0CTL1 |= UCSSEL_2 + UCRXEIE + UCBRKIE + UCDORM + UCSWRST;	//SMClock, UCRXEIE = 1, UCBRKEIE = 1; UCDORM = 1; UCTXADDR = 0, UCTXBRK = 0, UCSWRST = 1;

	UCA0BR0 = 125;
	UCA0BR1 = 0;
	UCA0MCTL = 0;    //UCOS16 = 0; Baud Rate = 128000, oversample mode disabled.

	UCA0STAT = 0x0;

//	IE2 |= UCA0RXIE;			//UCA0TXIE, UCA0RXIE = 01
//	IE2 |= UCA0TXIE;

	UCA0CTL1 &= ~UCSWRST;
}

void EnableUARTForReceive()
{
	IE2 |= UCA0RXIE;			//Enable Receive interruption
	IE2 &= ~UCA0TXIE;			//Disable Transmit interruption
}

void EnableUARTForTransmition()
{
	IE2 &= ~UCA0RXIE;
	IE2 |= UCA0TXIE;
}

void DisableUARTForReceiveOnly()
{
	IE2 &= ~UCA0RXIE;
}

void EnableUARTForReceiveOnly()
{
	IE2 |= UCA0RXIE;
}
void DisableUART()
{
	IE2 &= ~UCA0RXIE;			//Enable Receive interruption
	IE2 &= ~UCA0TXIE;			//Disable Transmit interruption
}

void UARTSendSingleByte(unsigned char Byte)
{
	DisableUARTForReceiveOnly();
    while(!(UCA0TXIFG & IFG2))
    {}
   	UCA0CTL1 |= UCTXADDR;
   	UCA0TXBUF = Byte;
    Delay10mS();
   	EnableUARTForReceiveOnly();
}

#pragma vector = USCIAB0RX_VECTOR
__interrupt void UCA0_RX_ISR(void)
{
	DISABLE_INT;
	static unsigned char UCA0Status = 0;
	static unsigned char ReceivedByte = 0;

	if (1 == BufferFull && COMMAND_PROCESSING == CommandStatus)
	{
		goto IntRet;
	}

	if(UCMODE_1 == (UCA0CTL0 & (UCMODE0 | UCMODE1)))		//idle line mode identification
	{

		UCA0Status = UCA0STAT;								//Any Receive Error, return

		if(UCRXERR & UCA0Status)
		{
			ReceivedByte = UCA0RXBUF;
			goto IntRet;
		}

		ReceivedByte = UCA0RXBUF;

		if (UCIDLE & UCA0Status)								//Address byte received
		{
			if(LOCAL_ADDRESS != ReceivedByte)
			{
				UCA0CTL1 |= UCDORM;
				goto IntRet;
			}

			if(UCDORM & ~UCA0CTL1)								//if UCDORM is already cleared, return;
			{
				ReceivedByte = UCA0RXBUF;
				goto IntRet;
			}

			if(POWER_MODE_SLEEP == PowerMode)					//if address is correct, then exit low power mode.
			{
				RESTORE_WATCHDOG;
				LPM3_EXIT;
				SleepTimer = 0;									//enforce SleepTimer to be 0, no need to acquire semaphore
				PowerMode = POWER_MODE_ACTIVE;
			}
			UCA0CTL1 &= ~UCDORM;								//Local Address Identified, prepare to receive following bytes.
			CommandStatus = COMMAND_RECEIVING;
			DataCounterInFrame = 0;
			ParseBufferIndex = ReceiveBufferIndex;
			UARTCounter = TickCounter;							//Keep UARTCounter synchronized to TickCounter;
		}else
		{
			if (ReceiveBufferIndex < BUFFER_SIZE)
			{
				ReceiveBuffer[ReceiveBufferIndex] = ReceivedByte;

				if (0 == DataCounterInFrame)
				{
					FrameLength = ReceivedByte;
				}

				ReceiveBufferIndex++;						//ReceiveBufferIndex should be decremented outside this interrupt routine.
				DataCounterInFrame++;
				UARTCounter = TickCounter;					//Keep UARTCounter synchronized to TickCounter;

			}else
			{
				BufferFull = 1;								//Clear BufferFull should also be done outside this routine.
			}
		}

		if (FrameLength + 2 == DataCounterInFrame && FrameLength != 0)	//check if a complete frame has been received.
			//The Frame for write command is like |Device Address|FrameLength|OperationType|Command1|Command2|...|Commandn|CheckSum|, FrameLength is the byte number from Operation to Commandn, start from 1
			//The Frame for receiving data is like |Device Address|FrameLength|OperationType|Command1|Command2|...|Commandn|ReadBytesCount|Checksum|, FrameLength includes one more byte for ReadByteCount.
		{
			FrameLength = 0;
			DataCounterInFrame = 0;
			UCA0CTL1 |= UCDORM;
			CommandStatus = COMMAND_PROCESSING;
			UARTCounter = 0;							//Clear UARTCounter after one frame has been received.
		}
	}

IntRet:
	UCA0Status = UCA0STAT;
	ENABLE_INT;
}

#pragma vector = USCIAB0TX_VECTOR
__interrupt void UCA0_TX_ISR(void)
{
    while(!(UCA0TXIFG & IFG2))
    {}
    if(SendByteNumber)
    {
    	UCA0TXBUF = SendBuffer[SendBufferIndex];
    	SendBufferIndex++;
    	SendByteNumber--;
    }else
    {
    	SendBufferIndex = 0;
    	SendBuffer = NULL;
    	DISABLE_UART_SEND_INT;
    }
}

#pragma vector = TIMER0_A0_VECTOR
__interrupt void Timer0_ISR(void)
{
	TickCounter++;

	P3OUT ^= BIT6;				//Test if the clock divider is correct, can be used as heart beat signal

	if(!(Semaphore & SLEEPTIMER_SEMAPHORE) && (-1 != SleepDelay) && (POWER_MODE_ACTIVE == PowerMode))
	{
		Semaphore |= SLEEPTIMER_SEMAPHORE;
		SleepTimer++;
		Semaphore &= ~SLEEPTIMER_SEMAPHORE;

		if(POWER_MODE_ACTIVE == PowerMode)			//Entering the Active Mode
		{
			if( SleepDelay <= SleepTimer)
			{
				SleepTimer = 0;
				PowerMode = POWER_MODE_SLEEP;
			}
		}
	}
}

void Delay1S()
{
	unsigned char i;

	for (i = 0; i < 100; i++)
	{
		Delay10mS();
	}

	WDT_CLEAR;
}

#pragma vector = PORT1_VECTOR
__interrupt void Port1_ISR(void)
{
	unsigned char DeviceStatus = 0, PortStatus = 0;

	if(P1IFG & BIT3)
	{
		P1IE &= ~BIT3;				//disable interrupt

		while(BIT3 & P1IFG)
		{
			P1IFG &= ~BIT3;			//clear interrupt flag
		}

		TURN_ON_DCDC;

		EnableUARTForReceive();
	}

	if(P1IFG & BIT4)				//alert from upper device
	{
		PortStatus = P4OUT;

		ENABLE_UPPER_BQ769x0;
		GetStatusRegister(UPPER_DEVICE, &DeviceStatus);
		ClearStatusRegister(UPPER_DEVICE, DeviceStatus);

		while(BIT4 & P1IFG)
		{
			P1IFG &= ~BIT4;
		}

		P4OUT = PortStatus;
	}

	if(P1IFG & BIT5)				//alert from bottom device
	{
		PortStatus = P4OUT;

		ENABLE_BOTTOM_BQ769x0;
		GetStatusRegister(BOTTOM_DEVICE, &DeviceStatus);
		ClearStatusRegister(BOTTOM_DEVICE, DeviceStatus);

		while(BIT5 & P1IFG)
		{
			P1IFG &= ~BIT5;
		}

		P4OUT = PortStatus;
	}
}

#pragma vector = PORT2_VECTOR
__interrupt void Port2_ISR(void)
{
	unsigned int SentByte, i;

	if(P2IFG & BIT3)
	{
		while(BIT3 & P2IFG)
		{
			P2IFG &= ~BIT3;
		}
		I2CInitialise_G2955();
		I2CSendBytes(BQ78350, (unsigned char*)ShutdownCommand, 3, &SentByte);							//send command to shutdown bq78350
		Delay10mS();
		Delay10mS();
		I2CSendBytes(BQ78350, (unsigned char*)ShutdownCommand, 3, &SentByte);							//send command twice to shutdown bq78350 in case the device is sealed
		while(1)
		{
		}
		//PowerMode = POWER_MODE_GOTO_SHUTDOWN;
		SleepTimer = 0;
	}

	if(BIT4 & P2IFG)
	{
		P2IE &= ~BIT4;			//disable interrupt for P2.4

		PowerMode = POWER_MODE_SLEEP_ACTIVE;		//Define a transition stage from SLEEP to ACTIVE
		SleepTimer = 0;
	}
		RESTORE_WATCHDOG;
		LPM3_EXIT;
}

int CheckFrameIntegrality()
{
	unsigned char LocalFrameLength = 0, CheckSum = 0;
	int i = 0;

	LocalFrameLength = ReceiveBuffer[FRAME_LENGTH_INDEX];

	for(i = 0; i < LocalFrameLength + 1; i++)
	{
		CheckSum += ReceiveBuffer[i];
	}

	if(CheckSum != ReceiveBuffer[ LocalFrameLength + 1])
	{
		return -1;
	}else
	{
		return 0;
	}
}

void UARTSendBytes(unsigned char *Buffer, unsigned char NumberOfBytes)
{
	if(NULL == SendBuffer)
	{
		SendBuffer = Buffer;
		SendByteNumber = NumberOfBytes;
		EnableUARTForTransmition();
	}
}

void FrameResponse(unsigned char FrameResponseCode)
{
	unsigned char i, *Data, ResponseCheckSum = 0;

	if(0 == FrameResponseCode || FrameResponseCode > 0x7f)
	{
		UARTSendSingleByte(FrameResponseCode);
		ResponseCheckSum += FrameResponseCode;
	}else
	{
		UARTSendSingleByte(FrameResponseCode);									//FrameResponseCode here can be the datalength of the data to be sent back to the host
		ResponseCheckSum += FrameResponseCode;
		UARTSendBytes(Command.CommandBody.ReadCommandBody.ReadData, Command.CommandBody.ReadCommandBody.ReadDataLength);
		Data = Command.CommandBody.ReadCommandBody.ReadData;
		while(SendByteNumber)
		{}

		for (i = 0; i < Command.CommandBody.ReadCommandBody.ReadDataLength; i++)
		{
			ResponseCheckSum += *Data;
			Data++;
		}

	}

	UARTSendSingleByte(ResponseCheckSum);
	EnableUARTForReceive();

	FrameLength = 0;
	DataCounterInFrame = 0;
	UCA0CTL1 |= UCDORM;
	CommandStatus = COMMAND_IDLE;
	UARTCounter = 0;							//Clear UARTCounter after one frame has been received.
	ReceiveBufferIndex = 0;
}

void ComposeCommand()
{
	int i = 0, result = -1;
    unsigned char CommandLength = 0;

    if(COMMAND_PROCESSING == CommandStatus && NO_PENDING_COMMAND == Command.Operation_Type)
    {

    	result = CheckFrameIntegrality();

        if (-1 == result)
        {
        	FrameResponse(FRAME_RESPONSE_CHECKSUM_ERROR);
        	return;
        }

    	CommandLength = ReceiveBuffer[ParseBufferIndex] - 1;
    	if(WRITE == ReceiveBuffer[ParseBufferIndex + 1])
    	{

    	}

    	if(EXECUTE == ReceiveBuffer[ParseBufferIndex + 1])
    	{
    		Command.CommandBody.CommandData = (unsigned char *)malloc(CommandLength);
    		Command.Operation_Type = EXECUTE;
    		Command.CommandLength = CommandLength;				//subtract 2 for bytes OperationType and Checksum
    		ParseBufferIndex += 2;
    		for(i = 0; i < Command.CommandLength; i++)
    		{
    			Command.CommandBody.CommandData[i] = ReceiveBuffer[ParseBufferIndex];
    			ParseBufferIndex++;
    		}
    	}

    	if(READ == ReceiveBuffer[ParseBufferIndex + 1])
    	{
 
    	}
    	ParseBufferIndex = 0;
    }
}

void ProcessCommand()
{
	unsigned int NumberOfBytes;
	int result = -1;

	if(NO_PENDING_COMMAND != Command.Operation_Type)
	{
		switch(Command.Operation_Type)
		{
		case(READ):

			break;
		case(WRITE):

			break;
		case(EXECUTE):
			switch(Command.CommandBody.CommandData[0])
			{
			case(BOTTOM_DSG_FET_ON):
				ENABLE_BOTTOM_BQ769x0;
				EnableDSG(BOTTOM_DEVICE);
				break;
			case(BOTTOM_DSG_FET_OFF):
				ENABLE_BOTTOM_BQ769x0;
				DisableDSG(BOTTOM_DEVICE);
				break;
			case(BOTTOM_CHG_FET_ON):
				ENABLE_BOTTOM_BQ769x0;
				EnableCHG(BOTTOM_DEVICE);
				break;
			case(BOTTOM_CHG_FET_OFF):
				ENABLE_BOTTOM_BQ769x0;
				DisableCHG(BOTTOM_DEVICE);
				break;
			case(UPPER_DSG_FET_ON):
				ENABLE_UPPER_BQ769x0;
				EnableDSG(UPPER_DEVICE);
				break;
			case(UPPER_DSG_FET_OFF):
				ENABLE_UPPER_BQ769x0;
				DisableDSG(UPPER_DEVICE);
				break;
			case(UPPER_CHG_FET_ON):
				ENABLE_UPPER_BQ769x0;
				EnableCHG(UPPER_DEVICE);
				break;
			case(UPPER_CHG_FET_OFF):
				ENABLE_UPPER_BQ769x0;
				DisableCHG(UPPER_DEVICE);
				break;
			}
			FrameResponse(Command.CommandBody.CommandData[0] + 0x80);
			free(Command.CommandBody.CommandData);
			Command.CommandBody.CommandData = NULL;
			break;
		default:
			break;
		}
	}
	CommandStatus = COMMAND_IDLE;
	Command.Operation_Type = NO_PENDING_COMMAND;
}

void CheckUARTTimeOut()
{
	DISABLE_INT;
	if (0 == UARTCounter || abs(UARTCounter - TickCounter) < 2 || 0 == TickCounter)
	{
		goto exit;
	}
	else
	{
		FrameResponse(FRAME_RESPNOSE_UART_TIMEOUT);
	}
exit:
	ENABLE_INT;
}

bool CheckIfI2CEnabled(unsigned char Device)
{
	switch(Device)
	{
	case BOTTOM_DEVICE:
		if((P4OUT & (BIT7 + BIT6 + BIT2)) != BOTTOM_BQ769x0_ENABLED)
			return false;
		break;
	case UPPER_DEVICE:
		if((P4OUT & (BIT7 + BIT6 + BIT2)) != UPPER_BQ769x0_ENABLED)
			return false;
		break;
	default:
		return false;
	}

	return true;
}

int EnableDSG(unsigned char Device)
{
	unsigned char Data;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);
	Data |= _DSG_ON;
	result = I2CWriteRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, Data);

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);	//check if the DSG_ENABLE bit has been set

	if (0 == (Data & _DSG_ON))
		return BIT_OPERATION_FAIL;

	return result;
}

int DisableDSG(unsigned char Device)
{
	unsigned char Data;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return -1;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);
	Data &= ~_DSG_ON;
	result = I2CWriteRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, Data);

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);	//check if the DSG_ENABLE bit has been cleared

	if (Data & _DSG_ON)
		return BIT_OPERATION_FAIL;

	return result;
}

int EnableCHG(unsigned char Device)
{
	unsigned char Data;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return -1;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);
	Data |= _CHG_ON;
	result = I2CWriteRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, Data);

	Delay10mS();
	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);	//check if the CHG_ENABLE bit has been set

	if (!(Data & _CHG_ON))
		return BIT_OPERATION_FAIL;

	return result;
}

int DisableCHG(unsigned char Device)
{
	unsigned char Data;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return -1;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);
	Data &= ~_CHG_ON;
	result = I2CWriteRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, Data);

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);	//check if the CHG_ENABLE bit has been cleared

	if (Data & _CHG_ON)
		return BIT_OPERATION_FAIL;

	return result;
}

int GetStatusRegister(unsigned char Device, unsigned char *Status)
{

	unsigned char Data = 0;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_STAT, &Data);

	*Status = Data;

	return result;

}

int GetControlRegister2(unsigned char Device, unsigned char *ControlRegsiterData)
{

	unsigned char Data = 0;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);

	*ControlRegsiterData = Data;

	return result;

}

int ClearStatusRegister(unsigned char Device, unsigned char Status)
{
	unsigned char Data = 0;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	result = I2CWriteRegisterByteWithCRC(BQMAXIMO, SYS_STAT, Status);

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_STAT, &Data);

	if((Data & Status) != 0)
		return BIT_OPERATION_FAIL;

	return result;
}

int SetCCOneShot(unsigned char Device)
{
	unsigned char Data = 0;
	int result = 0;

	if (false == CheckIfI2CEnabled(Device))
		return GENERAL_CALL_FAIL;

	result = I2CReadRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, &Data);

	Data |= _CC_ONESHOT;

	result = I2CWriteRegisterByteWithCRC(BQMAXIMO, SYS_CTRL2, Data);

	return result;
}

void DelayXms(unsigned int TensOfMilliSeconds)
{
	unsigned int i;

	for (i = 0; i < TensOfMilliSeconds; i++)
	{
		Delay10mS();
	}

}

void Delay250mS()
{
    DelayXms(25);
}

void ReadBatStackVoltage()
{
	unsigned int ADC_Data = 0;

	ENABLE_BAT_VOLT_DIVIDER;

	//Delay250mS();
	ADC10CTL0 |= REFON;
	ADC10CTL0 |= ADC10ON;
	ADC10CTL0 |= ENC;
	ADC10CTL0 |= ADC10SC;			//Start Converter

	while(ADC10CTL1 & ADC10BUSY)
	{}

	while (!(ADC10CTL0 & ADC10IFG))
	{}

	ADC_Data = ADC10MEM;

	ADC10CTL0 &= ~ADC10IFG;
	ADC10CTL0 &= ~ENC;
	ADC10CTL0 &= ~ADC10ON;
	ADC10CTL0 &= ~REFON;		//stop converter

	DISABLE_BAT_VOLT_DIVIDER;
}

int main(void)
{
	unsigned int SentByte;
	unsigned char UARTData = 0xaa;
	int Result;
	unsigned char UpperDeviceStatus, BottomDeviceStatus, BottomDeviceControlRegsiter;


    WDT_HOLD;	// Stop watchdog timer

    DISABLE_INT;

    ClockInitialise();

    I2CInitialise_G2955();

    UARTInitialise_G2955();

    TimerAInitialise();

    Port1Init();

    Port3Init();

    Port4Init();
    //Port2Init();

    //P1OUT |= 0x8;

    ADC10Init();

    ENABLE_INT;

	//P2OUT &= ~BIT2;					//enable DCDC
	//P2IE &= ~BIT4;					//Disable Interrupt for P2.4 for communication awaking;
    EnableUARTForReceive();

    //TimerAStart();

    WDT_START;

    Command.Operation_Type = 0;

   	ENABLE_BOTTOM_BQ769x0;

    InitialisebqMaximo_20S(BOTTOM_DEVICE);

    ENABLE_UPPER_BQ769x0;

    InitialisebqMaximo_20S(UPPER_DEVICE);

   	ENABLE_UPPER_BQ769x0;

   	GetStatusRegister(UPPER_DEVICE, &UpperDeviceStatus);

   	if (0 != UpperDeviceStatus)
   		ClearStatusRegister(UPPER_DEVICE, UpperDeviceStatus);

   	Result = EnableDSG(UPPER_DEVICE);		//normal test with DSG FET On
   	Result = EnableCHG(UPPER_DEVICE);		//normal test with CHG FET On

//   		Result = DisableDSG(UPPER_DEVICE);
//   		Result = DisableCHG(UPPER_DEVICE);		//for load removal test

   	ENABLE_BOTTOM_BQ769x0;

   	GetStatusRegister(BOTTOM_DEVICE, &BottomDeviceStatus);

   	if (0 != BottomDeviceStatus)
   		ClearStatusRegister(BOTTOM_DEVICE, BottomDeviceStatus);

   	Result = EnableDSG(BOTTOM_DEVICE);		//normal test with DSG FET On
  	Result = EnableCHG(BOTTOM_DEVICE);		//normal test with CHG FET On

//		Result = DisableDSG(UPPER_DEVICE);
//		Result = DisableCHG(UPPER_DEVICE);		//for load removal test

   	if (-1 == Result)
   		I2CInitialise_G2955();

	Delay10mS();
	//Delay10mS();

	//TURN_ON_DCDC;

    while(1)
    {
    	WDT_CLEAR;

//     	Result = UpdateVoltageFromBottomBq769x0();

//    	Result = UpdateVoltageFromUpperBq769x0();
//
    	Delay10mS();
    	Delay10mS();

       	ENABLE_UPPER_BQ769x0;

//       	UpdateVoltageFromBqMaximo();

       	GetStatusRegister(UPPER_DEVICE, &UpperDeviceStatus);

       	if (0 != (0x7F & UpperDeviceStatus))
       	{
           	Result = DisableDSG(UPPER_DEVICE);
           	Result = DisableCHG(UPPER_DEVICE);
       	}
//       	if (0 == (0x7F & UpperDeviceStatus))
//       	{
//           	Result = EnableDSG(UPPER_DEVICE);
//           	Result = EnableCHG(UPPER_DEVICE);
//       	}
//       	else
//       		ClearStatusRegister(UPPER_DEVICE, UpperDeviceStatus);

       	Result = CheckBqMaximoConfiguration_20S(UPPER_DEVICE);
       	if (-1 == Result)
       	{
           	Result = DisableDSG(UPPER_DEVICE);
           	Result = DisableCHG(UPPER_DEVICE);
       	}

       	ENABLE_BOTTOM_BQ769x0;

//       	UpdateVoltageFromBqMaximo();

       	GetStatusRegister(BOTTOM_DEVICE, &BottomDeviceStatus);

       	if (0 != (0x7F & UpperDeviceStatus))
       	{
           	Result = DisableDSG(UPPER_DEVICE);
           	Result = DisableCHG(UPPER_DEVICE);
       	}
//       	if (0 == (0x7F & BottomDeviceStatus))
//       	{
//           	Result = EnableDSG(BOTTOM_DEVICE);
//           	Result = EnableCHG(BOTTOM_DEVICE);
//       	}
//       	else
//       		ClearStatusRegister(BOTTOM_DEVICE, BottomDeviceStatus);

        Result = CheckBqMaximoConfiguration_20S(BOTTOM_DEVICE);

        if (-1 == Result)
        {
           	Result = DisableDSG(BOTTOM_DEVICE);
           	Result = DisableCHG(BOTTOM_DEVICE);
        }

    	if (COMMAND_PROCESSING == CommandStatus)				//this semaphore operation is not necessary, just follow conventional semaphore operation
    	{
    		ComposeCommand();
    		ProcessCommand();
    	}

 //Above is for short circuit test
/*

    	ReadBatStackVoltage();

    	TURN_ON_DCDC;

    	DelayXms(10);

    	TURN_OFF_DCDC;
*/
/*
    	ENABLE_BOTTOM_BQ769x0;
    	GetStatusRegister(BOTTOM_DEVICE, &BottomDeviceStatus);

       	if (0 != BottomDeviceStatus)
       		ClearStatusRegister(BOTTOM_DEVICE, BottomDeviceStatus);

       	SetCCOneShot(BOTTOM_DEVICE);

       	Delay250mS();
       	ClearStatusRegister(BOTTOM_DEVICE, _CC_READY);
       	GetStatusRegister(BOTTOM_DEVICE, &BottomDeviceStatus);
       	GetControlRegister2(BOTTOM_DEVICE, &BottomDeviceControlRegsiter);
*/
//       	ENABLE_EXTERNAL_REFERENCE;
/*
    	ENABLE_UPPER_BQ769x0;
    	GetStatusRegister(UPPER_DEVICE, &UpperDeviceStatus);

       	if (0 != UpperDeviceStatus)
       		ClearStatusRegister(BOTTOM_DEVICE, UpperDeviceStatus);

       	SetCCOneShot(UPPER_DEVICE);

       	Delay250mS();
       	ClearStatusRegister(UPPER_DEVICE, _CC_READY);
       	GetStatusRegister(UPPER_DEVICE, &BottomDeviceStatus);
       	GetControlRegister2(UPPER_DEVICE, &BottomDeviceControlRegsiter);
*/
//
//    	DISABLE_EXTERNAL_REFERENCE;
/*
    	ENABLE_UPPER_BQ769x0;
       	GetStatusRegister(UPPER_DEVICE, &UpperDeviceStatus);
       	if (0 != UpperDeviceStatus)
       		ClearStatusRegister(UPPER_DEVICE, UpperDeviceStatus);
       	Result = DisableDSG(UPPER_DEVICE);
       	Result = DisableCHG(UPPER_DEVICE);

    	ENABLE_BOTTOM_BQ769x0;
       	GetStatusRegister(BOTTOM_DEVICE, &BottomDeviceStatus);
       	if (0 != BottomDeviceStatus)
       		ClearStatusRegister(BOTTOM_DEVICE, BottomDeviceStatus);
       	Result = DisableDSG(BOTTOM_DEVICE);
       	Result = DisableCHG(BOTTOM_DEVICE);

       	ENABLE_UPPER_BQ769x0;
       	GetStatusRegister(UPPER_DEVICE, &UpperDeviceStatus);
       	if (0 != UpperDeviceStatus)
       		ClearStatusRegister(UPPER_DEVICE, UpperDeviceStatus);
       	Result = EnableDSG(UPPER_DEVICE);
       	Result = EnableCHG(UPPER_DEVICE);

       	ENABLE_BOTTOM_BQ769x0;
       	GetStatusRegister(BOTTOM_DEVICE, &BottomDeviceStatus);
       	if (0 != BottomDeviceStatus)
       		ClearStatusRegister(BOTTOM_DEVICE, BottomDeviceStatus);
       	Result = EnableDSG(BOTTOM_DEVICE);
       	Result = EnableCHG(BOTTOM_DEVICE);
*/
 //This section flip-flops the DSG and CHG driving signal so as to observe the waveform of DSG and CHG driving signals
//    	CheckUARTTimeOut();
    	/*
    	if (COMMAND_PROCESSING == CommandStatus)				//this semaphore operation is not necessary, just follow conventional semaphore operation
    	{
    		if(!(Semaphore & SLEEPTIMER_SEMAPHORE))
    		{
    			Semaphore |= SLEEPTIMER_SEMAPHORE;
    			SleepTimer = 0;
    			Semaphore &= ~SLEEPTIMER_SEMAPHORE;
    		}
    		ComposeCommand();
    		ProcessCommand();
    	}

		if (POWER_MODE_SLEEP == PowerMode)
		{
			P1OUT |= BIT4;
			//I2CSendBytes(BQ78350, (unsigned char*)SleepCommand, 3, &SentByte);	//bq78350 enter sleep mode
			DISABLE_INT;
			//PullDownSMBus();
			DisableUART();
			P2OUT |= BIT2;						//disable DCDC output
			//DISABLE_INT;
			P2IE |= BIT4;						//enable interrupt for P2.4
			while(P2IFG & BIT4)
			{
				P2IFG &= ~BIT4;
			}									//just in case of successive pulse causes entering the P2_ISR interrupt routine here, once this happens,
												//the P2.4 interrupt will be disabled under LPM3 mode in the interrupt routine,
												//this causes the device never returns to active mode
			ENTER_LPM3;							//enter LPM3, meanwhile, the interrupt is enabled.
		}

		if (POWER_MODE_SLEEP_ACTIVE == PowerMode)
		{
			if(P2OUT & BIT2)
				P2OUT &= ~BIT2;					//Enable DCDC output
			//P2IE &= ~BIT4;						//Disable Interrupt for P2.4, this should be done in interrupt routine;
			//ReviveSMBus();
			I2CInitialise_G2955();
			Delay1S();
			FrameResponse(FRAME_RESPONSE_AWAKENED);
			PowerMode = POWER_MODE_ACTIVE;
		}*/
    }

    return Result;
}
